import asyncio
import signal
import socket

import grpc
import yaml
from concurrent import futures

from etcd_register.etcd.client import EtcdClient
from etcd_register.util.config import Config


class Rpc(object):
    etcd = None

    def register(self, filepath, function, instance, message_size=205109840):
        """

        :param filepath: 配置文件
        :param function: 注册函数
        :param instance: 注册实例
        :param message_size: 单次请求数据大小
        :return:
        """
        config = Config.get_config(filepath)
        addr = ":{}".format(config.get("etcd"))
        if not addr:
            raise Exception("etcd地址: {} 不规范".format(addr))
        port = ":{}".format(config.get("port"), 12306)
        service = config.get("service")
        self.etcd = EtcdClient(addr)

        loop = asyncio.get_event_loop()
        loop.add_signal_handler(signal.SIGTERM, self.term_sig_handler, loop)
        loop.add_signal_handler(signal.SIGINT, self.term_sig_handler, loop)
        loop.run_until_complete(self.run(config, service, port, function, instance, message_size))

    async def run(self, cfg, service, port, function, instance, message_size):
        await asyncio.gather(
            Rpc.listen(function, instance, port, message_size),
            self.serve(cfg, service, instance, port),
        )

    def term_sig_handler(self, service, port, lp):
        self.etcd.unregister_service(service, Rpc.get_ip_address() + port).send(None)
        lp.stop()

    @staticmethod
    async def listen(function, instance, port, message_size):
        MAX_MESSAGE_LENGTH = message_size
        server = grpc.server(futures.ThreadPoolExecutor(max_workers=10),
                             options=[
                                 ('grpc.max_send_message_length', MAX_MESSAGE_LENGTH),
                                 ('grpc.max_receive_message_length', MAX_MESSAGE_LENGTH),
                             ]
                             )
        function(instance, server)
        server.add_insecure_port('[::]{}'.format(port))
        print("服务启动成功, 端口: ", port)
        server.start()
        await Rpc.wait(60)

    @staticmethod
    async def wait(timeout):
        while True:
            await asyncio.sleep(timeout)

    async def serve(self, cfg, service, instance, port):
        await self.etcd.register_api(service, instance, cfg)
        await self.etcd.register_service(service, Rpc.get_ip_address() + port, 10)

    @staticmethod
    def get_ip_address():
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.connect(("8.8.8.8", 80))
        return s.getsockname()[0]


reg = Rpc()



import yaml


class Config(object):
    @staticmethod
    def get_config(cfg):
        """
        获取配置文件内容
        :param cfg:
        :return:
        """
        try:
            with open(cfg, 'r', encoding='UTF-8') as f:
                data = yaml.load(f)
            return data
        except Exception as e:
            raise Exception("读取服务配置文件失败, error: {}".format(str(e)))



from etcd_register.util.config import Config
import requests


class HttpClient(object):
    conf = None

    def __init__(self, config):
        self.conf = Config.get_config(config)

    def invoke(self, service, method, json=None, version='v1'):
        try:
            url = self.conf.get('gateway')
            r = requests.post('{}/{}/{}/{}'.format(url, version, service, method), json=json)
            if r.status_code != 200:
                raise Exception("status code != 200")
            return r.json()
        except Exception as e:
            return dict(code=101, msg=str(e))



import asyncio
import json

from aioetcd3.client import client


class EtcdClient(object):
    client = None
    scheme = "ashe"

    def __init__(self, host):
        """
        :param host: ip:port,ip:port
        """
        self.client = client(host)

    async def unregister_service(self, name, addr):
        print("unregister service: ", "/{}/{}/{}".format(self.scheme, name, addr))
        await self.client.delete("/{}/{}/{}".format(self.scheme, name, addr))

    async def register_service(self, name, addr, ttl):
        while True:
            value, meta = await self.client.get("/{}/{}/{}".format(self.scheme, name, addr))
            if value is None:
                await self.with_alive(name, addr, ttl)
            await asyncio.sleep(ttl)

    @staticmethod
    def lower_first(s):
        if len(s) == 0:
            return ""
        if s[0].islower():
            if len(s) == 1:
                return s[0].lower()
            return s[0].lower() + s[1:]
        return ""

    async def register_api(self, name, instance, cfg):
        version = cfg.get("version")
        methods = cfg.get("method", {})
        for d in dir(instance):
            if d.startswith("_") or d.endswith("_"):
                continue
            if d not in methods.keys():
                print("方法: {}注册失败, 请在service.yml中配置".format(d))
                continue
            info = methods.get(d)
            await self.register_single(version, name, d, info)

    async def register_single(self, version, service, method_name, no_auth=None):
        key = "{}.{}.{}".format(version, EtcdClient.lower_first(service), EtcdClient.lower_first(method_name))
        info = {"no_auth": False if no_auth is None else no_auth.get("no_auth"),
                "path": "/{}/{}".format(service, method_name)}
        await self.client.put(key, json.dumps(info, ensure_ascii=False))

    async def with_alive(self, name, addr, ttl):
        lease = await self.client.grant_lease(ttl)
        key = "/{}/{}/{}".format(self.scheme, name, addr)
        print("service alive: {}".format(key))
        await self.client.put(key, addr, lease=lease)
        await self.refresh_lease(lease, ttl)

    async def refresh_lease(self, lease, ttl):
        try:
            while True:
                await self.client.refresh_lease(lease)
                await asyncio.sleep(ttl - 5)
        except Exception as err:
            print("服务续租失败, error: ", err)



import json

import grpc

from etcd_register.proto import rpc_pb2, rpc_pb2_grpc
from etcd_register.util.config import Config


class RpcClient(object):
    conf = None

    def __init__(self, config):
        self.conf = Config.get_config(config)

    @staticmethod
    def encode(data):
        return json.dumps(data).encode('utf-8')

    @staticmethod
    def decode(data):
        return data.decode('utf-8')

    def invoke(self, service, method, args=None, version='v1', origin=False):
        try:
            addr = self.conf.get('rpc')
            channel = grpc.insecure_channel(addr)
            stub = rpc_pb2_grpc.RpcServiceStub(channel)
            request = rpc_pb2.Args()
            request.version = version
            request.service = service
            request.method = method
            if isinstance(args, bytes):
                request.args = args
            else:
                request.args = json.dumps(args, ensure_ascii=False).encode('utf-8')
            response = stub.Invoke(request)
            if origin:
                return Response(response.code, response.resultJson, response.msg)
            data = response.resultJson.decode('utf-8')
            return Response(response.code, data, response.msg)
        except Exception as e:
            return Response(101, None, str(e))


class Response(object):
    code = None
    data = None
    msg = None

    def __init__(self, code, data, msg):
        self.code = code
        self.data = data
        self.msg = msg

    def marshal(self):
        return {"code": self.code, "data": self.data, "msg": self.msg}

    def __str__(self):
        return "code: {}\nmsg: {}\ndata: {}".format(self.code, self.msg, self.data)